#include <stdio.h>
#include <stdlib.h>
#include "graph_int.h"

graph_t *
graph_create()
{
  graph_t *g = (graph_t *)calloc(1, sizeof(graph_t));
  g->head = 0;
  g->headpi = 0;
  g->edgeno = 0;

  return g; 
}

void 
graph_free(graph_t *g)
{
  graph_freeedges(g);
  free(g);
}

void
graph_freeedges(graph_t *g)
{
  edge_t *item, *tmp;

  item = g->head;
  while (item) {
    tmp = item;
    item = tmp->next;
    free(tmp->info);
    free(tmp);
  }
  g->head = 0;
  g->headpi = 0;
  g->edgeno = 0;
}

void
graph_append_edge(graph_t *g, edge_t *edge)
{
  edge->prev = 0;
  if (g->head == 0) {
    edge->next = 0;
  } else {
    edge->next = g->head;
    g->head->prev = edge;
  }
  g->head = edge;
}

void
graph_append_edgepi(graph_t *g, edge_t *edge)
{
  edge->prevpi = 0;
  if (g->headpi == 0) {
    edge->nextpi = 0;
  } else {
    edge->nextpi = g->headpi;
    g->headpi->prevpi = edge;
  }
  g->headpi = edge;
}

edge_t *
graph_make_edge(unsigned int sn, unsigned int en, void *info)
{
  edge_t *edge = (edge_t *)calloc(1, sizeof(edge_t));
  
  edge->sn = sn;
  edge->en = en;
  edge->info = info;
  edge->prevpi = 0;
  edge->nextpi = 0;
  edge->nextadj = 0;
  edge->prevadj = 0;

  return edge;
}

edge_t *
graph_add_edge(graph_t *g, unsigned int sn, 
	       unsigned int en, void *info)
{
  edge_t *edge;
  
  edge = graph_make_edge(sn, en, info);
  graph_append_edge(g, edge);
  g->edgeno++;
  return edge;
}

void
graph_adjacent(edge_t **Adj, edge_t *edge)
{
  if (Adj[edge->sn] == 0)
    edge->nextadj = 0;
  else
    edge->nextadj = Adj[edge->sn];
  Adj[edge->sn] = edge;
}

void
graph_adjacent_en(edge_t **Adj, edge_t *edge)
{
  if (Adj[edge->en] == 0)
    edge->prevadj = 0;
  else
    edge->prevadj = Adj[edge->en];
  Adj[edge->en] = edge;
}

void
graph_adjacent_pi(edge_t **piAdj, edge_t *edge)
{
  if (piAdj[edge->sn] == 0)
    edge->nextpi = 0;
  else
    edge->nextpi = piAdj[edge->sn];
  piAdj[edge->sn] =  edge;
}

void 
set_make(unsigned int *p, unsigned int *rank, unsigned int x)
{
  p[x] = x;
  rank[x] = 0;
}

unsigned int 
set_find(unsigned int *p, unsigned int x)
{
  if (x != p[x])
    p[x] = set_find(p, p[x]);
  return p[x];
}

void 
set_link(unsigned int *p, unsigned int *rank, unsigned int x, 
	 unsigned int y)
{
  if (rank[x] > rank[y]) {
    p[y] = x;
  } else {
    p[x] = y;
    if (rank[x] == rank[y])
      rank[y]++;
  }
}

void 
set_union(unsigned int *p, unsigned int *rank, unsigned int x, 
	  unsigned int y)
{
  set_link(p, rank, set_find(p, x), set_find(p, y));
}

void no_mem_exit()
{
  printf("No memory.\n");
  exit(1);
}





