#include <stdio.h>
#include <stdlib.h>
#include "list_int.h"

list_t *
list_create(void)
{
  list_t *list = (list_t *)calloc(1, sizeof(list_t));

  list->num = 0;
  list->head = list->tail = 0;
  return list;
}


void
list_free(list_t *list)
{
  list_item *item, *tmp;

  item = list->head;
  while (item) {
    tmp = item;
    item = tmp->next;
    free(tmp);
  }
  free(list);
}

void
list_freeitems(list_t *list)
{
  list_item *item, *tmp;

  item = list->head;
  while (item) {
    tmp = item;
    item = tmp->next;
    free(tmp);
  }
  list->head = list->tail = 0;
  list->num = 0;
}	

void
list_append(list_t *list, void *data)
{
  list_item *item = (list_item *)malloc(sizeof(list_item));

  item->data = data;
  item->next = 0;

  if (list->tail) {
    list->tail->next = item;
    item->prev = list->tail;
  } else {
    list->head = item;
    item->prev = 0;
  }
  list->tail = item;
  list->num ++;
}

void
list_append2head(list_t *list, void *data)
{
  list_item *item = (list_item *)calloc(1, sizeof(list_item));

  item->data = data;
  item->prev = 0;

  if (list->head) {
    list->head->prev = item;
    item->next = list->head;
  } else {
    list->tail = item;
    item->next = 0;
  }
  list->head = item;
  list->num ++;
}

void
list_append_before(list_t *list, list_item *item, void *data)
{
  if (!item) 
    list_append(list, data);
  else if (item == list->head)
    list_append2head(list, data);
  else {
    list_item *newitem = (list_item *)calloc(1, sizeof(list_item));
    newitem->data = data;
    
    newitem->next = item;
    newitem->prev = item->prev;
    item->prev->next = newitem;
    item->prev = newitem;
    list->num ++;
  }
}

/******
The current implementation is based on bubble sort.
******/
void
list_bubble_sort(list_t *list, compare_fcn compare)
{
  int swapped = 1;
  list_item *item;

  while (swapped) {
    swapped = 0;
    item = list->head;
    while (item && item->next) {
      if (compare(item->data, item->next->data) <= 0) {
	item = item->next;
      } else {
	swapped = 1;
	if (item->prev) {
	  item->prev->next = item->next;
	} else {
	  list->head = item->next;
	}
	if (item->next->next) {
	  item->next->next->prev = item;
	} else {
	  list->tail = item;
	}
	item->next->prev = item->prev;
	item->prev = item->next;
	item->next = item->next->next;
	item->prev->next = item;
      }
    }
  }
}
  

/******
The current implementation is based on quick sort.
******/
void
exchange(list_t *list, list_item *i, list_item *temp)
{
  list_item *A, *B;
  A = i->prev;
  B = i->next;

  if (temp->next!=0) temp->next->prev = i;
  else  list->tail = i;
  i->next = temp->next;
  if (B!=temp) 
    {
      temp->prev->next = i;
      i->prev = temp->prev;
    }
  else   
    {     
      temp->next = i;
      i->prev = temp;
    }
  if (A!=0) A->next = temp;
  else list->head = temp;
  temp->prev = A;
  if (temp->next!=i) 
    {
      temp->next = B;
      B->prev = temp;     
    }
}


list_item *
partition(list_t *list, compare_fcn compare, list_item *head, list_item *tail)
{
  list_item *i, *j, *k, *temp;
  i = head; j = tail; 
  while ((i!=j) && (i->next!=j)) {
    i = i->next;
    j = j->prev;
  }
  temp = i;
  i = head; j = tail;
  if (compare(i->data, temp->data) * compare(i->data, j->data)<0)
    temp = i;
  else if (compare(j->data, temp->data) * compare(j->data, i->data)<0)
    temp = j;  
  if (i!=temp) exchange(list, i, temp);
 
  if (temp==j) j = i; 
  i = temp;

  do
    { while(compare(j->data, temp->data)>=0 && j!=i)
        j = j->prev;
      if (i!=j)
        { 
	  exchange(list, i, j);
          k = j; j = i; i = k->next;          
	}
      while(compare(i->data, temp->data)<=0 && i!=j)
	i = i->next;
      if (i!=j)
	{
          exchange(list, i, j);
          k = j; j = i->prev; i = k;         
	}
    }while(i!=j);
  return(i);
}



void
list_quick_sort(list_t *list, compare_fcn compare, list_item *head, list_item *tail)
{
  list_item *m, *h, *t;
  h = head->prev; t = tail->next;
  if ((head==tail)||(tail->next==head)) return;
  m = partition(list, compare, head, tail);
  while (head->prev!=h) head = head->prev;
  if (m!=head) list_quick_sort(list, compare, head, m->prev);
  while (tail->next!=t) tail = tail->next;
  if (m!=tail) list_quick_sort(list, compare, m->next, tail);
}
    
void list_sort(list_t *list, compare_fcn compare)
{
  list_quick_sort(list, compare, _list_head(list), _list_tail(list));
}

unsigned int
list_length(list_t *list)
{
  return list->num;
}

void *
list_extracthead(list_t *list)
{
  void *tmp;
  list_item *item;
  
  item = list->head;
  tmp = item->data;
  list->head = list->head->next;
  if (list->head) { 
    free(list->head->prev);
    list->head->prev = 0;
  } else {
      free(item); 
      list->tail = 0;
  }

  list->num --;
  return tmp;
}

void *
list_deleteitem(list_t *list, list_item *item)
{
  void *tmp;  
  
  tmp = item->data;
  if (item->prev) { 
    if (item->next) {
      item->prev->next = item->next;
      item->next->prev = item->prev;
    }
    else {
      item->prev->next = 0;
      list->tail = item->prev;
    }
  }
  else {
    if (item->next) {
      item->next->prev = 0;
      list->head = item->next;
    }
    else 
      list->head = list->tail = 0;
  }
  
  free(item);
  list->num --;
  return tmp;
}



list_item *
_list_head(list_t *list)
{
  return list->head;
}

list_item *
_list_tail(list_t *list)
{
  return list->tail;
}

void *
_list_data(list_item *item)
{
  return item->data;
}


list_item *
_list_next(list_item *item)
{
  return item->next;
}


