#include <stdio.h>
#include <stdlib.h>
#include "mcrp_int.h"

//#define PRINT_CYCLE
//#define DEBUG

void
mcrp_restore(graph_t *E, float *wbak)
{
  edge_t *edge;
  unsigned int i = 0;

  edge = E->head;
  while (edge) {
    ((info_t *)edge->info)->w = wbak[i++];
    edge = edge->next;
  }
}

void 
mcrp_reverse_bfs(graph_t *E, edge_t *edge, unsigned int *update,
		 edge_t **pi, float lamda, float *d)
{
  if (update[edge->en] == 0) 
    mcrp_reverse_bfs(E, pi[edge->en], update, pi, lamda, d);
  d[edge->sn] = d[edge->en]+((info_t *)edge->info)->w 
    -(((info_t *)edge->info)->t*lamda);
  update[edge->sn] = 1;
}

void 
mcrp_compute(unsigned int numnode, graph_t *E, float *lamda, 
	     unsigned int maxormin)
{
  float *d, *wbak, delay, tmplamda, delta;
  unsigned int *p, *rank, *update, i, flag = 1, n, s = 0, 
    improved, numedge, u, v;
  edge_t *edge, **pi;

  numedge = E->edgeno;
  if ((d = (float *)calloc(numnode+1, sizeof(float))) == NULL) no_mem_exit();
  if ((wbak = (float *)calloc(numedge, sizeof(float))) == NULL) no_mem_exit();
  if ((p = (unsigned int *)calloc(numnode+1, sizeof(unsigned int))) == NULL) no_mem_exit();
  if ((rank = (unsigned int *)calloc(numnode+1, sizeof(unsigned int))) == NULL) no_mem_exit();
  if ((update = (unsigned int *)calloc(numnode+1, sizeof(unsigned int))) == NULL) no_mem_exit();
  if ((pi = (edge_t **)calloc(numnode+1, sizeof(edge_t *))) == NULL) no_mem_exit();

  for (i=0; i<=numnode; i++) {
    d[i]=P_INFINITY;
    p[i]=0;
  }
  i = 0;
  edge = E->head;
  while (edge) {
    // Backup w for later restore.
    wbak[i++] = ((info_t *)edge->info)->w;
    u = edge->sn;
    if (maxormin == 0) 
      ((info_t *)edge->info)->w = -1 * ((info_t *)edge->info)->w;
    if (((info_t *)edge->info)->w < d[u]) {
      d[u] = ((info_t *)edge->info)->w;
      pi[u] = edge;
    }
    edge = edge->next;
  }

  while (flag==1) {
    tmplamda = P_INFINITY;
    edge = E->head;
    for(i=0; i<=numnode; i++)
      set_make(p, rank, i);
    while (edge) {
      if (pi[edge->sn] == edge) 
        graph_append_edgepi(E, edge); 
      edge = edge->next;
    } 
    
    // Examine every cycle. 
    edge = E->headpi;
    while (edge) {
      u = edge->sn;
      v = edge->en;
      if (set_find(p, u) != set_find(p, v)) 
        set_union(p, rank, u, v);
      else {
        delay = 0; 
	n = 0; 
        do {
          delay += ((info_t *)edge->info)->w;
          n += ((info_t *)edge->info)->t;
          edge = pi[edge->en];
        } while(edge->sn != u);
        if (n == 0) {
          printf("Zero FF Cycle!!!:\n");
          edge = pi[u];
	  do {
	    printf("(%d,%d w=%f,t=%d)\t", edge->sn, edge->en,
		   -((info_t *)edge->info)->w, 
		   ((info_t *)edge->info)->t);
	    edge = pi[edge->en];
	  } while (edge->sn != u);
	  printf("\n\n");
	  exit(1);
        } 
	else {
          if (delay/n < tmplamda) {
            tmplamda = delay/n;
            if (maxormin && (delay < 0)) {
              *lamda = delay;
#ifdef PRINT_CYCLE
	      u=edge->sn;printf("\nFind a negative cycle:\n");
	      do {
		printf("%d->%d:%f\t", edge->sn,edge->en,((info_t *)edge->info)->w);
		edge = pi[edge->en];
	      } while (edge->sn!=u); 
#endif
	      free(d);
	      free(wbak);
	      free(p);
	      free(rank);
	      free(update);
	      free(pi);
	      return;
	    }
	    s = edge->sn;
          }
        } 
#ifdef DEBUG	
        printf("\n Current lamda:%f",delay/n);
        printf("\n Correspond cycle:\n");
        edge1=pi[u];
	do {
	   printf("(%d,%d)",edge1->sn,edge1->en);
	   edge1=pi[edge1->en];
	}while (edge1->sn!=u);
#endif
      }
      edge = edge->nextpi;
    }

    if (tmplamda == P_INFINITY) {
      *lamda = tmplamda;
      free(d);
      free(wbak);
      free(p);
      free(rank);
      free(update);
      free(pi);
      return;
    }

    // Reverse BFS.
#ifdef DEBUG
    printf("%d, %f\n",s,tmplamda);
#endif
    for (i=0; i<=numnode; i++) 
      update[i] = 0;
    update[s] = 1;  
    edge = E->headpi;
    while(edge){
      if ((update[edge->sn] == 0) &&
	  (set_find(p, edge->sn) == set_find(p, s))) 
	mcrp_reverse_bfs(E, edge, update, pi, tmplamda, d);
      edge = edge->nextpi;
    }
    E->headpi = 0;

    // Improve node distances.
    improved = 0;
    edge = E->head;
    while(edge) {
      u = edge->sn;
      v = edge->en;
      delta = d[u] - (d[v] + ((info_t *)edge->info)->w 
		      - (((info_t *)edge->info)->t*tmplamda));
      if (delta > EPSI) { 
        improved = 1;
        d[u] = d[v] + ((info_t *)edge->info)->w 
	  - (((info_t *)edge->info)->t*tmplamda);
        pi[edge->sn] = edge;
      }
      edge = edge->next;
    }
 
    if (improved == 0) {
      flag = 0;
      if (maxormin == 0)
	*lamda = -1 * tmplamda;
      else *lamda = tmplamda;
#ifdef PRINT_CYCLE   
      // Output max ratio cycle.
      printf("Cycle: ");
      edge=pi[s];
      do {
        printf("(%d,%d w=%f,t=%d)\t", 
	       edge->sn, edge->en, -((info_t *)edge->info)->w, 
	       ((info_t *)edge->info)->t);
        edge=pi[edge->en];
      }while (edge->sn != s);
#endif
    }
  }  

  mcrp_restore(E, wbak);
  free(d);
  free(wbak);
  free(p);
  free(rank);
  free(update);
  free(pi);
}








