#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "retiming.h"

void 
delete_extra_edge(graph_t *E, graph_t *hE, unsigned int numnode)
{
  edge_t *edge, *tmp=0;
  edge = E->head;
  E->edgeno = 0;
  while ((edge->sn != numnode+1) && (edge->en != numnode+1)) {
    tmp = edge;
    E->edgeno ++;
    edge = edge->next;
  }
  if (tmp)
    tmp->next = 0;
  hE->head = edge;
}

unsigned int
check_reducing_loop(unsigned int numnode, unsigned int u, 
		    unsigned int *pi)
{
  unsigned int *count, i, flag = 1;
  if ((count = (unsigned int *)calloc(numnode+1, sizeof(unsigned int))) == NULL) no_mem_exit();	
   
  for (i=1; i<=numnode; i++) 
    count[i] = 0;	  
  count[u] = 1;
  i = u;
  while ((pi[i]!=0) && (count[i]<2)) {
    i = pi[i];
    count[i] ++;
  }
  if (i == u)
    flag = 0;
  if (!flag)
    printf("Infeasible due to neg CYCLE found!!!\n");
  free(count);
  return flag;
}

void
reset_pi(unsigned int u, edge_t **Adj, edge_t **revAdj, 
	 unsigned int *pi)
{
  edge_t *edge;

  edge = Adj[u];
  while (edge) {
     if (pi[edge->en] == u)
       pi[edge->en] = 0;
     edge = edge->nextadj;    
  }
  edge = revAdj[u];
  while (edge) {
    if (((info_t *)edge->info)->p==0 && pi[edge->sn]==u)
      pi[edge->sn] = 0;
    edge = edge->prevadj;
  }
}	

unsigned int
retiming_compute(unsigned int numnode, graph_t *E, 
		 graph_t *hE, float lamda, edge_t **Adj, 
		 edge_t **revAdj, float *fd, float *bd, 
		 int maxFF, int mode, int *r)
{
  unsigned int num, i, u, v, numedge, flag = 1, io, *pi;
  float *R, midv1, midv2, min, newr;
  edge_t *edge;
  list_t *list = list_create();
  int *rcomp;
  long j;

  num = numnode + 1;
  numedge = E->edgeno;
  if ((R = (float *)calloc(num, sizeof(float))) == NULL) no_mem_exit();
  if ((pi = (unsigned int *)calloc(num, sizeof(unsigned int))) == NULL) no_mem_exit();
  if ((rcomp = (int *)calloc(num, sizeof(int))) == NULL) no_mem_exit();

  for (i=1; i<=numnode; i++) {
    R[i] = P_INFINITY;   // R()=r()+t()/T
    r[i] = P_INFINITY;   // retiming tag
    pi[i] = 0;           // predecessor pointer
    rcomp[i]=0;          // for comparison use
  }
  edge = hE->head;
  while (edge) {
    if (edge->en == numnode+1) {
      io = edge->sn;
      R[io] = 1;
      r[io] = 0;
      list_append(list, (void *)io);
    }
    edge = edge->next;
  }
  
  j = 0;	  
  while ((list_length(list)>0) && (flag)) {
    u = (unsigned int)list_extracthead(list);
    j = (j+1)%(numnode*numnode);
    if (j == 1) {
      i = 1;
      while (flag && (r[i]==rcomp[i]) && (i<=numnode))
        i++;
      if (i > numnode) {
        flag = 0;
        printf("Infeasible: no r change within |V| iterations.\n");
      }
      else {
        for (i=1;i<=numnode;i++)
          rcomp[i] = r[i];
      }
    }

    edge = Adj[u];
    while (edge && flag) {
      // Edge (u,v).
      v = edge->en;
      if (mode && (((info_t *)edge->info)->p==0) 
	  && (((info_t *)edge->info)->t>0)) {
        if (r[v] > r[u]) {
          r[v] = r[u];
          R[v] = r[v] + 1;
          pi[v] = u;
          reset_pi(v, Adj, revAdj, pi);
	  list_append(list, (void *)v);
	}
      } 
      else {
	midv1 = R[u]+((info_t *)edge->info)->t
	  -((info_t *)edge->info)->w/lamda; 
	midv2 = midv1-fd[edge->en]/lamda;
	if (midv2-floor(midv2) > 1)
	  midv2 = floor(midv2)+1;
	midv2 = floor(midv2)+1-bd[edge->en]/lamda;
	min = midv1 < midv2 ? midv1 : midv2;	  
	newr = midv1-fd[edge->en]/lamda;
	if (newr-floor(newr) > 1)
	  newr = floor(newr) + 1;
	else 
	  newr=floor(newr);
	if (R[v] > min) {
	  pi[v] = u;
	  if (mode) {
	    if (R[v]-r[v] == min-newr)
	      flag = check_reducing_loop(numnode, v, pi);
	    if (newr < -maxFF) {
	      // for (i=1;i<=numnode;i++)
	      //   if (r[i]<-maxFF) printf("r[%d]=%d\n",i,r[i]);
	      flag=0;
	      printf("Infeasible due to r<-maxFF!!!\n");
	    }
	  }
	  if (flag) {
	    R[v] = min;
	    r[v] = newr;
	    reset_pi(v, Adj, revAdj, pi);
	    list_append(list, (void *)v);
	  }
	}
      }
      edge = edge->nextadj;
    }

    edge = revAdj[u];    
    while (mode && edge && flag) {
      // Edge (v,u).
      v = edge->sn;
      if ((((info_t *)edge->info)->p == 0) && (r[v] > r[u])) {
          r[v] = r[u];
          R[v] = r[v] + 1 - bd[v]/lamda;
          pi[v] = u;  
          reset_pi(v, Adj, revAdj, pi);
	  list_append(list, (void *)v);
      }	 
      edge = edge->prevadj;
    }
  }  
 
  edge = hE->head;
  while (edge && flag) {
    if ((edge->sn == numnode+1) && (r[edge->en] < 0)) {
      flag = 0;
      printf("Infeasible due to neg PIPO PATH found!!!\n");
    }
    edge = edge->next;    
  } 

  free(R);
  free(pi);
  free(rcomp);
  list_free(list);
/*
  if (flag) {
    for (i=1;i<=numnode;i++)
      printf("r[%d]=%d\n",i,r[i]);
  }
  if (mode) 
    printf("No.of updates in current iteration= %d\n",N);
*/

  return flag;  
}
